#!/bin/sh

checkout_casadi () {
    if !(test -d "casadi"); then
        echo Checking out casadi from github
        git clone --recursive https://github.com/casadi/casadi.git -b master casadi
    else
        echo casadi already checked out
    fi
}

build_casadi () {
    checkout_casadi
    # Build casadi libraries and install in /usr/local, 
    # does not build the python extension
    BUILD_NAME=casadi/build

    mkdir -p $BUILD_NAME
    cd $BUILD_NAME
    cmake -DWITH_COMMON=ON \
      -DWITH_BUILD_DSDP=ON -DWITH_DSDP=ON \
      -DWITH_OSQP=ON -DWITH_QPOASES=ON ..

    make -j$(sysctl -n hw.physicalcpu)

    sudo make install
    cd ..
    rm -rf $BUILD_NAME
}

build_casadi_python () {
    checkout_casadi    
    # Build python extension
    # build_casadi_python python_interpreter abbreviation

    PYTHON_INTERPRETER=$1
    PYVER=$(sed -e "s/^.*3\.*\([0-9]*\)$/3\1/g" <<< $PYTHON_INTERPRETER)
    ENV_NAME=env_build_$PYVER
    BUILD_NAME=casadi/build_$PYVER
    LIB_NAME=lib_$PYVER
    TAR_NAME=casadi_$PYVER.tar.gz
    $PYTHON_INTERPRETER -m venv $ENV_NAME
    
    source $ENV_NAME/bin/activate
    pip install -U pip setuptools wheel
    pip install ipython
    mkdir -p $BUILD_NAME
    cd $BUILD_NAME
    cmake -DWITH_PYTHON=ON -DWITH_PYTHON3=ON -DWITH_COMMON=ON \
      -DWITH_BUILD_DSDP=ON -DWITH_DSDP=ON \
      -DWITH_OSQP=ON -DWITH_QPOASES=ON \
      -DPYTHON_PREFIX=../../$LIB_NAME ..

    make -j$(sysctl -n hw.physicalcpu) python

    make install_python
    cd ../../$LIB_NAME
    tar -cf ../$TAR_NAME casadi

    cd ..
    rm -rf $BUILD_NAME $LIB_NAME
    deactivate
    rm -rf $ENV_NAME

    echo "Build completed. Unpack $TAR_NAME and add to your PYTHONPATH."
}

if [[ ! -z $1 ]] && [[ $1 = "--build-lib" ]]; then
    echo Building casadi library and installing in /usr/local
    build_casadi
elif [[ ! -z $1 ]] && [[ $1 = "--install-prerequisites" ]]; then
    if [ -z $(which brew) ] ; then 
        echo "Installing prerequisites requires homebrew package manager (https://brew.sh)"
    else
        echo Installing pre-requisites
        brew install wget swig gcc cmake pkg-config
        brew install ipopt
    fi
elif [[ ! -z $1 ]] && [[ ! -z $2 ]] && [[ $1 = "--build-python" ]]; then
    echo Building Python extensions
     build_casadi_python $2
else
    echo $0 "[--install-prerequisites | --build-lib | --build-python interpreter]"
    echo
    echo "--install-prerequisites     Installs required packages using homebrew (wget swig gcc cmake pkg-config ipopt)"
    echo "--build-lib                 Builds casadi library and installs in /usr/local (requires sudo rights)"
    echo "--build-python interpreter  Builds python API for interpreter"
    echo
    echo "Call build script in this order to build all."
fi
